import arrow
import zipfile
from .structures import PBO
from construct import Container

def find_files_index_by_filename(pbo, fn):
        i = 0
        for header in pbo.header:
                if header.meta:
                        continue
                if header.filename == fn:
                        return i
                i += 1
        raise RuntimeError("Can't locate target file within pbo")


def find_header_index_by_filename(pbo, fn):
        for i, header in enumerate(pbo.header):
                if header.meta:
                        continue
                if header.filename == fn:
                        return i
        raise RuntimeError("Can't locate target file within pbo")


def pbo_empty():
    arrow_z = arrow.Arrow(1970, 1, 1, 0, 0, 0)
    pbo = PBO.parse(PBO.build(Container(
        fields=Container(value=Container(
            header=[
                Container(filename=u'.', method=0, originalsize=0, timestamp=arrow_z, datasize=0, meta=False, final=False, extended=False, extension=None),
                Container(filename=u'', method=0, originalsize=0, timestamp=arrow_z, datasize=0, meta=True, final=True, extended=False, extension=None)
            ],
            files=[b''],
        ))
    )))
    del(pbo.files[0])
    del(pbo.header[0])

# extented header
#    pbo.header.insert(0, Container(filename=u'', method=0, originalsize=0, timestamp=arrow_z, datasize=0, meta=True, final=False, extended=True, extension=['foo','bar','']))

    return pbo


def pbo_files_add(pbo, fn, data):
    arrow_z = arrow.Arrow(1970, 1, 1, 0, 0, 0)
    f = 0
    h = 0
    for header in pbo.header:
        if header.final:
            break
        if header.extended:
            h += 1
            continue
        if fn.lower() == header.filename.lower():
            raise RuntimeError('file already exists in pbo')
        if fn.lower() < header.filename.lower():
            break
        f += 1
        h += 1
    pbo.header.insert(h, Container(filename=fn, method=0, originalsize=0, timestamp=arrow_z, datasize=len(data), meta=False, final=False, extended=False, extension=None))
    pbo.files.insert(f, data)


def find_indexes_by_filename(pbo, fn):
    i = 0
    for h, header in enumerate(pbo.header):
        if header.meta:
            continue
        if header.filename == fn:
            return i, h
        i += 1
    raise RuntimeError("Can't locate target file within pbo")

def pbo_remove_file(pbo, fn):
    fi, hi = find_indexes_by_filename(pbo, fn)
    del(pbo.files[fi])
    del(pbo.header[hi])

def pbo_by_zip(fn):
    import zipfile
    pbo = pbo_empty()
    zf = zipfile.ZipFile(fn)
    # find all files and strip first component
    filelist = (('/'.join(fn.split('/')[1:]).replace('/', '\\'), fn) for fn in zf.namelist() if not fn.endswith('/'))
    for fn, zfn in filelist:
        pbo_files_add(pbo, fn, zf.read(zfn))
    return pbo



import os
import os.path as path
import zipfile
from .structures import PBO
from .utils import pbo_empty, pbo_files_add, find_files_index_by_filename

class PBOFile():
    def __init__(self, init_empty=True):
        if init_empty:
            self._pbo = pbo_empty()

    @classmethod
    def from_bytes(cls, data):
        """create PBOFile from data"""
        obj = PBOFile(init_empty=False)
        obj._pbo = PBO.parse(data)
        return obj

    @classmethod
    def from_file(cls, fn):
        """read .pbo file"""
        obj = PBOFile(init_empty=False)
        obj._pbo = PBO.parse_file(fn)
        return obj

    def to_file(self, fn):
        """write .pbo file"""
        open(fn, 'wb').write(self.as_bytes())

    def as_bytes(self):
        """build .pbo file bytes"""
        return PBO.build(self._pbo)

    @classmethod
    def from_zip(cls, fn):
        pbo = PBOFile()
        zf = zipfile.ZipFile(fn)
        # find all files and strip first component
        filelist = (('/'.join(fn.split('/')[1:]), fn) for fn in zf.namelist() if not fn.endswith('/'))
        for fn, zfn in filelist:
            pbo.add_file_data(fn, zf.read(zfn))
        return pbo

    def to_zip(self, fn):
        pass

    @classmethod
    def from_directory(cls, root):
        # maybe use .pboignore?
        os.chdir(root)
        fns = [os.path.join(dirpath, name)[2:] for dirpath, dirnames, filenames in os.walk('.') for name in filenames]
        fns.sort(key=lambda x:x.lower())

        obj = PBOFile()
        for fn in fns:
            obj.add_file(fn, fn)
        return obj

    def to_directory(self, root):
        for pbofn in self.filenames():
            *dirs, fn = pbofn.split('\\')
            os.makedirs(path.join(root, *dirs), exist_ok=True)
            fullfn = path.join(root, *dirs, fn)
            open(fullfn, 'wb').write(self.file_as_bytes(pbofn))

    def filenames(self):
        for h in self._pbo.header:
            if h.meta:
                continue
            yield(h.filename)

    def add_file(self, fn, data_fn):
        fn = fn.replace('/', '\\')
        pbo_files_add(self._pbo, fn, open(data_fn,'rb').read())

    def add_file_data(self, fn, data):
        fn = fn.replace('/', '\\')
        pbo_files_add(self._pbo, fn, data)

    def remove_file(self, fn):
        fn = fn.replace('/', '\\')
        pbo_remove_file(self._pbo, fn)

    def file_as_bytes(self, fn):
        fn = fn.replace('/', '\\')
        i = find_files_index_by_filename(self._pbo, fn)
        return self._pbo.files[i]



from construct import *
from datetime import datetime
import arrow
import hashlib

class TimestampAdapter(Adapter):
    def _decode(self, obj, context, path):
        return datetime.utcfromtimestamp(obj)

    def _encode(self, obj, context, path):
        if obj == datetime(1970, 1, 1, 0, 0):
            return 0
        else:
            return int(obj.timestamp())


# https://resources.bisimulations.com/wiki/PBO_File_Format

PBO_Header_Extended = RepeatUntil(lambda obj, lst, ctx: len(obj) == 0, CString("ascii"))

PBO_Header_Entry = Struct(
    "filename" / Default(CString("ascii"), ""),
    "method" / Enum(Int32ul, default=0, packed=0x43707273, extended=0x56657273) * "PackingMethod",
    "originalsize" / Int32ul * "OriginalSize",
    Const(0, Int32ul) * "Reserved",
    "timestamp" / Timestamp(Int32ul, 1, 1970) * "TimeStamp",  # or use TimestampAdapter(Int32ul) if you don't like Arrow,
    "datasize" / Int32ul * "DataSize",

    "i" / Computed(lambda ctx: ctx._index),
    "meta" / Computed(lambda ctx: len(ctx.filename) == 0),
    "final" / Computed(lambda ctx: ctx.i != 0 and len(ctx.filename) == 0),
    "extended" / Computed(lambda ctx: ctx.i == 0 and len(ctx.filename) == 0),

    "extension" / If(this.extended, PBO_Header_Extended),
)


class Files(Construct):
    def _parse(self, stream, context, path):
        obj = ListContainer()

        for header in context.header:
            if header.meta:
                continue
            if header.method.intvalue != 0:
                raise RuntimeError("can't extract files with method: 0x%x" % (header.method.intvalue,))
            e = Bytes(header.datasize)._parsereport(stream, context, path)
            obj.append(e)

        return obj

    def _build(self, obj, stream, context, path):
        retlist = ListContainer()
        i = 0
        for header in context.header:
            if header.meta:
                continue
            context._index = i
            buildret = Bytes(header.datasize)._build(obj[i], stream, context, path)
            retlist.append(buildret)
            i += 1
        return retlist


class PBOStruct(Struct):
    def _parse(self, stream, context, path):
        return super()._parse(stream, context, path)

    def _build(self, obj, stream, context, path):
        # adjust file sizes
        i = 0
        for header in obj.fields.value.header:
            if header.meta:
                continue
            header.datasize = len(obj.fields.value.files[i])
            i += 1

        filecount = len(obj.fields.value.files)

        if filecount != i:
            raise RuntimeError("not all files were processed, expected %d, processed %d" % (i, filecount))

        del(obj.fields['data'])

        return super()._build(obj, stream, context, path)


PBO = PBOStruct(
    "fields" / RawCopy(Struct(
        "header" / RepeatUntil(lambda obj, lst, ctx: obj.final is True, PBO_Header_Entry),
        "files" / Files(),
    )),
    "data_end" / Tell,
    Const(0, Byte) * "terminator",
    "sha1" / Checksum(Bytes(20), lambda data: hashlib.sha1(data).digest(), this.fields.data),
    Terminated,

    "header" / Computed(this.fields.value.header),
    "files" / Computed(this.fields.value.files),
)



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='pboutil',
      version='0.1',
      description='Read and write ArmA PBO files.',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://gitlab.com/arma3-server-tools/pboutil',
      author='Taneli Kaivola',
      author_email='dist@ihme.org',
      license='GPLv3',
      packages=['pboutil'],
      install_requires=[
          'construct >=2.9, <2.10',
          'arrow',
      ],
      classifiers=(
          "Development Status :: 4 - Beta",
          "Environment :: Console",
          "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
          "Operating System :: OS Independent",
          "Programming Language :: Python :: 3",
          "Topic :: Games/Entertainment :: First Person Shooters",
          "Topic :: System :: Archiving",
      ),
      zip_safe=True)




from construct import *
from pprint import pprint
from datetime import datetime
import arrow
import hashlib
from .pbofile import PBOFile
from .structures import PBO, Files
from .utils import find_files_index_by_filename, find_header_index_by_filename, find_indexes_by_filename, \
                   pbo_empty, pbo_files_add, pbo_remove_file, pbo_by_zip
